{ unit MixerAPI

  This unit uses multimedia audio functions to control speaker output, which is needed
  for the applications run on Windows XP.

  Author : Silhwan Hyun  (hyunsh@hanafos.com)
  Contributor : Emil Weiss (weiss.em@arcor.de)

  #### Copyright Notice ####
  Copyright 2011 Silhwan Hyun, All Rights Reserved
  This unit is free. It may be used both in commercial and non-commercial software either in
  original or in modified form.
  This unit can be freely distributed in any way and by any means provided this copyright
  notice is preserved.

  Revision History
  ------------------------------
  v1.0.0           (25 Feb 2017)
  - fixed some parts of MixerApi
  (by Silhwan Hyun)

  v0.9.8           (5 Feb 2017)
  - Added function SetMixerBalance
  (by Emil Weiss)

  v0.9.1           (13 Mar 2012)
  - Added a parameter to function GetComponentControl, to support multi sound cards in a system.

  v0.9.0           (3 Aug 2011)
  - Initial release

}

unit MixerAPI;

interface

uses Windows, SysUtils, MMSystem, Math;

const
  MaxControl = 8;
  MaxChannel = 8;

type
  TVolume = array [0 .. MaxChannel - 1] of DWORD;

  TDestLineInfo = record
    MixerId: DWORD;
    DestType: DWORD;
    LineName: string;
    Channels: DWORD;
    MuteControl: TMixerControl;
    VolumeControl: TMixerControl;
  end;

function GetComponentControl(Mixer: hMixerObj; ComponentType: DWORD;
  var DestLineInfo: TDestLineInfo): MMResult;
function IsMute(Mixer: hMixerObj; Control_Mute: TMixerControl): Boolean;
function MuteLine(Mixer: hMixerObj; Control_Mute: TMixerControl; Mute: Boolean): MMResult;
function GetMixerVolume(Mixer: hMixerObj; Channels: DWORD; Control_Vol: TMixerControl;
  var Volume: TVolume): MMResult;
function SetMixerVolume(Mixer: hMixerObj; Channels: DWORD; Control_Vol: TMixerControl;
  Volume: TVolume): MMResult;
function SetMixerBalance(Mixer: hMixerObj; Channels: DWORD; Control_Vol: TMixerControl;
  Balance: Integer; Balance_Max: Integer; Balance_Mid: Integer): MMResult;

implementation

// Check ReturnCode and display the error information if ReturnCode is not MMSYSERR_NOERROR.
procedure CheckRetStat(ReturnCode: MMResult);
var
  StatusText: string;
begin
  if ReturnCode = MMSYSERR_NOERROR then
    exit
  else if ReturnCode = MMSYSERR_ERROR then
    StatusText := 'Unspecified error'
  else if ReturnCode = MMSYSERR_ALLOCATED then
    StatusText := 'Already allocated'
  else if ReturnCode = MIXERR_INVALCONTROL then
    StatusText := 'Invalid Mixer Control'
  else if ReturnCode = MIXERR_INVALLINE then
    StatusText := 'Invalid Mixer Line'
  else if ReturnCode = MMSYSERR_BADDEVICEID then
    StatusText := 'Bad Device Id'
  else if ReturnCode = MMSYSERR_INVALFLAG then
    StatusText := 'Invalid Flag'
  else if ReturnCode = MMSYSERR_INVALHANDLE then
    StatusText := 'Invalid HANDLE'
  else if ReturnCode = MMSYSERR_INVALPARAM then
    StatusText := 'Invalid PARAMETER'
  else if ReturnCode = MMSYSERR_NODRIVER then
    StatusText := 'No Mixer device is available for the object specified by hmxob'
  else if ReturnCode = MMSYSERR_NOMEM then
    StatusText := 'Unable to allocate or lock memory'
  else
    StatusText := 'Unknown Error';

  if ReturnCode <> MMSYSERR_NOERROR then
    MessageBox(0, PChar(StatusText), 'Confirm', mb_OK);
end;

// Get Mixer Line info, Mute control and Volume control of specified component type.
function GetMixerControl(Mixer: hMixerObj; ComponentType: DWORD; var MixerLine: TMixerLine;
  var Control_Mute, Control_Vol: TMixerControl): MMResult;
var
  Controls: TMixerLineControls;
begin
  ZeroMemory(@MixerLine, SizeOf(MixerLine));
  MixerLine.cbStruct := SizeOf(MixerLine);
  MixerLine.dwComponentType := ComponentType;
  Result := mixerGetLineInfo(Mixer, @MixerLine, MIXER_GETLINEINFOF_COMPONENTTYPE);

  if Result = MMSYSERR_NOERROR then
  begin
    ZeroMemory(@Controls, SizeOf(Controls));
    Controls.cbStruct := SizeOf(Controls);
    Controls.dwLineID := MixerLine.dwLineID;
    Controls.cControls := 1;
    Controls.dwControlType := MIXERCONTROL_CONTROLTYPE_MUTE;
    Controls.cbmxctrl := SizeOf(Control_Mute);
    Controls.pamxctrl := @Control_Mute;
    Result := mixerGetLineControls(Mixer, @Controls, MIXER_GETLINECONTROLSF_ONEBYTYPE);
  end;

  if Result = MMSYSERR_NOERROR then
  begin
    ZeroMemory(@Controls, SizeOf(Controls));
    Controls.cbStruct := SizeOf(Controls);
    Controls.dwLineID := MixerLine.dwLineID;
    Controls.cControls := 1;
    Controls.dwControlType := MIXERCONTROL_CONTROLTYPE_VOLUME;
    Controls.cbmxctrl := SizeOf(Control_Vol);
    Controls.pamxctrl := @Control_Vol;
    Result := mixerGetLineControls(Mixer, @Controls, MIXER_GETLINECONTROLSF_ONEBYTYPE);
  end;
end;

// Get info of specified mixer component type among mixer destination lines.
function GetComponentControl(Mixer: hMixerObj; ComponentType: DWORD;
  var DestLineInfo: TDestLineInfo): MMResult;
var
  MixerLine: TMixerLine;
  Control_Mute, Control_Vol: TMixerControl;
  pName: PChar;
begin

  Result := GetMixerControl(Mixer, ComponentType, MixerLine, Control_Mute, Control_Vol);

  if Result = MMSYSERR_NOERROR then
  begin
    DestLineInfo.MixerId := Mixer;
    DestLineInfo.DestType := ComponentType;
    pName := @MixerLine.szName[0];
    DestLineInfo.LineName := pName;
    DestLineInfo.Channels := MixerLine.cChannels;
    DestLineInfo.MuteControl := Control_Mute;
    DestLineInfo.VolumeControl := Control_Vol;
  end;
end;

// Get muting state of specified mute control. (off = true)
function IsMute(Mixer: hMixerObj; Control_Mute: TMixerControl): Boolean;
var
  ControlDetails: TMixerControlDetails;
  RetStat: MMResult;
  BoolValue: array [0 .. MaxChannel - 1] of longbool;
begin

  ControlDetails.cbStruct := SizeOf(ControlDetails);
  ControlDetails.dwControlId := Control_Mute.dwControlId;
  ControlDetails.cMultipleItems := 0;
  ControlDetails.cbDetails := 4; // = the length of MIXERCONTROLDETAILS_BOOLEAN
  ControlDetails.cChannels := 1; // assume MIXERCONTROL_CONTROLF_UNIFORM
  ControlDetails.paDetails := @BoolValue;

  RetStat := mixerGetControlDetails(Mixer, @ControlDetails,
    MIXER_GETCONTROLDETAILSF_VALUE + MIXER_OBJECTF_MIXER);

  CheckRetStat(RetStat);

  if RetStat = MMSYSERR_NOERROR then
    if BoolValue[0] = False then
      Result := False
    else
      Result := true
    else
      Result := False;
end;

// Set muting state of specified mute control.
function MuteLine(Mixer: hMixerObj; Control_Mute: TMixerControl; Mute: Boolean): MMResult;
var
  ControlDetails: TMixerControlDetails;
  BoolValue: array [0 .. MaxChannel - 1] of longbool;
  i: Integer;
begin
  if Mute then
    i := 1
  else
    i := 0;

  BoolValue[0] := BOOL(i); // ** We should use very tricky code like this.
  ControlDetails.cbStruct := SizeOf(ControlDetails);
  ControlDetails.dwControlId := Control_Mute.dwControlId;
  ControlDetails.cMultipleItems := 0;
  ControlDetails.cbDetails := 4; // = the length of MIXERCONTROLDETAILS_BOOLEAN
  ControlDetails.cChannels := 1; // assume MIXERCONTROL_CONTROLF_UNIFORM
  ControlDetails.paDetails := @BoolValue;

  Result := mixerSetControlDetails(Mixer, @ControlDetails,
    MIXER_SETCONTROLDETAILSF_VALUE + MIXER_OBJECTF_MIXER);

  CheckRetStat(Result);
end;

// Get volume level of specified volume control.
function GetMixerVolume(Mixer: hMixerObj; Channels: DWORD; Control_Vol: TMixerControl;
  var Volume: TVolume): MMResult;
var
  ControlDetails: TMixerControlDetails;
  IsUniformControl: Boolean;
begin
  IsUniformControl := ((Control_Vol.fdwControl and MIXERCONTROL_CONTROLF_UNIFORM)
      = MIXERCONTROL_CONTROLF_UNIFORM);

  ControlDetails.cbStruct := SizeOf(ControlDetails);
  ControlDetails.dwControlId := Control_Vol.dwControlId;

  if not IsUniformControl then
    ControlDetails.cChannels := Channels
  else
    ControlDetails.cChannels := 1;

  ControlDetails.cMultipleItems := 0;
  ControlDetails.cbDetails := 4; // = the length of MIXERCONTROLDETAILS_UNSIGNED
  ControlDetails.paDetails := @Volume;

  Result := mixerGetControlDetails(Mixer, @ControlDetails,
    MIXER_GETCONTROLDETAILSF_VALUE + MIXER_OBJECTF_MIXER);
  CheckRetStat(Result);
end;

// Set volume level of specified volume control.
function SetMixerVolume(Mixer: hMixerObj; Channels: DWORD; Control_Vol: TMixerControl;
  Volume: TVolume): MMResult;
var
  ControlDetails: TMixerControlDetails;
  IsUniformControl: Boolean;
begin
  IsUniformControl := ((Control_Vol.fdwControl and MIXERCONTROL_CONTROLF_UNIFORM)
      = MIXERCONTROL_CONTROLF_UNIFORM);

  ControlDetails.cbStruct := SizeOf(ControlDetails);
  ControlDetails.dwControlId := Control_Vol.dwControlId;

  if not IsUniformControl then
    ControlDetails.cChannels := Channels
  else
    ControlDetails.cChannels := 1;

  ControlDetails.cMultipleItems := 0;
  ControlDetails.cbDetails := 4; // = the length of MIXERCONTROLDETAILS_UNSIGNED
  ControlDetails.paDetails := @Volume;

  Result := mixerSetControlDetails(Mixer, @ControlDetails,
    MIXER_SETCONTROLDETAILSF_VALUE + MIXER_OBJECTF_MIXER);
  CheckRetStat(Result);
end;

function SetMixerBalance(Mixer: hMixerObj; Channels: DWORD; Control_Vol: TMixerControl;
  Balance: Integer; Balance_Max: Integer; Balance_Mid: Integer): MMResult;
var
  V: Integer;
  Volume: TVolume;
  IsUniformControl: Boolean;

begin
  Result := MMSYSERR_ERROR;

  if Channels = 1 then
    exit;

  IsUniformControl := ((Control_Vol.fdwControl and MIXERCONTROL_CONTROLF_UNIFORM)
      = MIXERCONTROL_CONTROLF_UNIFORM);
  if IsUniformControl then
    exit;

  if GetMixerVolume(Mixer, Channels, Control_Vol, Volume) <> MMSYSERR_NOERROR then
    exit;

  V := Max(Volume[0], Volume[1]);
  if Balance < Balance_Mid then // left sided balance
  begin
    Volume[0] := V;
    Volume[1] := V * Balance div Balance_Mid
  end
  else if Balance = Balance_Mid then // center balance
  begin
    Volume[0] := V;
    Volume[1] := V;
  end
  else // right sided balance
  begin
    Volume[0] := V * (Balance_Max - Balance) div Balance_Mid;
    Volume[1] := V;
  end;

  Result := SetMixerVolume(Mixer, Channels, Control_Vol, Volume);
end;

end.
